@@ -17,12 +17,15 @@
 
 
 
+(setq cl-user:*al-compile-when-defining* t)
+
+
+
 (in-package cl)
 
 
 
-
-(defparameter *features* '(:common-lisp))
+(defparameter *features* '(:alisp :common-lisp :unix))
 
 
 
@@ -73,6 +76,93 @@
 
 
 
+(defun first (list) (nth 0 list))
+(defun second (list) (nth 1 list))
+(defun third (list) (nth 2 list))
+(defun fourth (list) (nth 3 list))
+(defun fifth (list) (nth 4 list))
+(defun sixth (list) (nth 5 list))
+(defun seventh (list) (nth 6 list))
+(defun eighth (list) (nth 7 list))
+(defun ninth (list) (nth 8 list))
+(defun tenth (list) (nth 9 list))
+
+
+(defun rest (list) (cdr list))
+
+
+(defun caar (list) (car (car list)))
+(defun cadr (list) (nth 1 list))
+(defun cdar (list) (cdr (car list)))
+(defun cddr (list) (nthcdr 2 list))
+(defun caaar (list) (car (car (car list))))
+(defun caadr (list) (car (nth 1 list)))
+(defun cadar (list) (nth 1 (car list)))
+(defun caddr (list) (car (nthcdr 2 list)))
+(defun cdaar (list) (cdr (car (car list))))
+(defun cdadr (list) (cdr (car (cdr list))))
+(defun cddar (list) (nthcdr 2 (car list)))
+(defun cdddr (list) (nthcdr 3 list))
+(defun caaaar (list) (car (car (car (car list)))))
+(defun caaadr (list) (car (car (nth 1 list))))
+(defun caadar (list) (car (nth 1 (car list))))
+(defun caaddr (list) (car (nth 1 (cdr list))))
+(defun cadaar (list) (nth 1 (car (car list))))
+(defun cadadr (list) (nth 1 (nth 1 list)))
+(defun caddar (list) (nth 1 (cdr (car list))))
+(defun cadddr (list) (car (nthcdr 3 list)))
+(defun cdaaar (list) (cdr (car (car (car list)))))
+(defun cdaadr (list) (cdr (car (car (cdr list)))))
+(defun cdadar (list) (cdr (car (cdr (car list)))))
+(defun cdaddr (list) (cdr (car (nthcdr 2 list))))
+(defun cddaar (list) (nthcdr 2 (car (car list))))
+(defun cddadr (list) (nthcdr 2 (nth 1 list)))
+(defun cdddar (list) (nthcdr 3 (car list)))
+(defun cddddr (list) (nthcdr 4 list))
+
+
+
+(defmacro cond (&body body)
+  (let (ret l)
+    (dolist (f body)
+      (if (cdr f)
+	  (progn
+	    (let ((cl
+		   `(if ,(car f)
+			,(if (cddr f)
+			     `(progn ,@(cdr f))
+			     (if (cdr f)
+				 (cadr f)))
+			,nil)))
+	      (if ret
+		  (setf (car l) cl l (cdddr cl))
+		  (setf ret cl l (cdddr cl)))))
+	  (progn
+	    (let* ((s (gensym))
+		   (cl
+		    `(let ((,s ,(car f)))
+		       (if ,s
+			   ,s
+			   ,nil))))
+	      (if ret
+		  (setf (car l) cl l (cddr (cdaddr cl)))
+		  (setf ret cl l (cddr (cdaddr cl))))))))
+    ret))
+
+
+
+(defmacro while (cond &rest body)
+  (let ((tag (gensym)))
+    `(tagbody
+	,tag
+	(if ,cond
+	    (progn
+	      (progn
+		,@body)
+	      (go ,tag))))))
+
+
+
 (defun identity (x) x)
 
 (defun constantly (x) (lambda (&rest r) x))
@@ -272,61 +362,17 @@
 
 (defun gentemp (&optional (pr "T") (pack *package*))
   (let ((str (make-string-output-stream)))
-    (loop
-     (write-string pr str)
-     (write gentemp-counter :stream str)
-     (incf gentemp-counter)
-     (let ((s (get-output-stream-string str)))
-       (if (not (find-symbol s pack))
-	   (return-from gentemp (values (intern s pack))))))))
+    (do nil
+	(nil)
+      (write-string pr str)
+      (write gentemp-counter :stream str)
+      (setq gentemp-counter (+ 1 gentemp-counter))
+      (let ((s (get-output-stream-string str)))
+	(if (not (find-symbol s pack))
+	    (return-from gentemp (values (intern s pack))))))))
 
 
 
-(defun first (list) (nth 0 list))
-(defun second (list) (nth 1 list))
-(defun third (list) (nth 2 list))
-(defun fourth (list) (nth 3 list))
-(defun fifth (list) (nth 4 list))
-(defun sixth (list) (nth 5 list))
-(defun seventh (list) (nth 6 list))
-(defun eighth (list) (nth 7 list))
-(defun ninth (list) (nth 8 list))
-(defun tenth (list) (nth 9 list))
-
-
-(defun rest (list) (cdr list))
-
-
-(defun caar (list) (car (car list)))
-(defun cadr (list) (nth 1 list))
-(defun cdar (list) (cdr (car list)))
-(defun cddr (list) (nthcdr 2 list))
-(defun caaar (list) (car (car (car list))))
-(defun caadr (list) (car (nth 1 list)))
-(defun cadar (list) (nth 1 (car list)))
-(defun caddr (list) (car (nthcdr 2 list)))
-(defun cdaar (list) (cdr (car (car list))))
-(defun cdadr (list) (cdr (car (cdr list))))
-(defun cddar (list) (nthcdr 2 (car list)))
-(defun cdddr (list) (nthcdr 3 list))
-(defun caaaar (list) (car (car (car (car list)))))
-(defun caaadr (list) (car (car (nth 1 list))))
-(defun caadar (list) (car (nth 1 (car list))))
-(defun caaddr (list) (car (nth 1 (cdr list))))
-(defun cadaar (list) (nth 1 (car (car list))))
-(defun cadadr (list) (nth 1 (nth 1 list)))
-(defun caddar (list) (nth 1 (cdr (car list))))
-(defun cadddr (list) (car (nthcdr 3 list)))
-(defun cdaaar (list) (cdr (car (car (car list)))))
-(defun cdaadr (list) (cdr (car (car (cdr list)))))
-(defun cdadar (list) (cdr (car (cdr (car list)))))
-(defun cdaddr (list) (cdr (car (nthcdr 2 list))))
-(defun cddaar (list) (nthcdr 2 (car (car list))))
-(defun cddadr (list) (nthcdr 2 (nth 1 list)))
-(defun cdddar (list) (nthcdr 3 (car list)))
-(defun cddddr (list) (nthcdr 4 list))
-
-
 (defun (setf caar) (newval list)
   (let ((c (car list)))
     (rplaca c newval)
@@ -551,11 +597,36 @@
 
 
 (defun (setf subseq) (newval seq start &optional end)
-  (let ((len (- (or end (length seq)) start))
-	(newvallen (length newval)))
-    (dotimes (i (min len newvallen))
-      (setf (elt seq (+ i start)) (elt newval i)))
-    newval))
+  (if (consp seq)
+      (let ((i 0)
+	    (cons newval))
+	(setq seq (nthcdr start seq))
+	(do nil
+	    (nil)
+	  (if (or (not seq)
+		  (not cons)
+		  (and (not (consp newval)) (>= i (length newval)))
+		  (and end (>= i (- end start))))
+	      (return-from nil newval))
+	  (rplaca seq (if (consp newval)
+			  (car cons)
+			  (elt newval i)))
+	  (setq seq (cdr seq))
+	  (if (consp newval)
+	      (setq cons (cdr cons)))
+	  (setq i (+ 1 i))))
+      (let ((len (- (or end (length seq)) start))
+	    (cons newval))
+	(dotimes (i len)
+	  (if (or (not cons)
+		  (and (not (consp newval)) (>= i (length newval))))
+	      (return-from nil newval))
+	  (setf (elt seq (+ i start)) (if (consp newval)
+					  (car cons)
+					  (elt newval i)))
+	  (if (consp newval)
+	      (setq cons (cdr cons))))
+	newval)))
 
 
 
@@ -652,19 +723,25 @@
 
 
 (defmacro define-modify-macro (name lambda-list func &optional doc)
-  (let (args)
-    (dolist (l lambda-list)
-      (unless (eq l '&optional)
-	(if (typep l 'cons)
-	    (setq args (cons (car l) args))
-	    (setq args (cons l args)))))
+  (let (args restvar)
+    (do ((l lambda-list (cdr l)))
+	((not l))
+      (unless (eq (car l) '&optional)
+	(if (typep (car l) 'cons)
+	    (setq args (cons (caar l) args))
+	    (if (eq (car l) '&rest)
+		(progn
+		  (setq restvar (cadr l))
+		  (setq l (cdr l)))
+		(setq args (cons (car l) args))))))
+    (setq args (reverse args))
     `(defmacro ,name (place . ,lambda-list)
        (let ((exp (multiple-value-list (get-setf-expansion place)))
 	     letf)
 	 (dolist (f (car exp))
 	   (setq letf (cons (list f (caadr exp)) letf))
 	   (setf (cadr exp) (cdadr exp)))
-	 (setq letf (cons (list (caaddr exp) (list ',func (nth 4 exp) ,@args)) letf))
+	 (setq letf (cons (list (caaddr exp) (,(if restvar 'list* 'list) ',func (nth 4 exp) ,@args ,@(if restvar (list restvar)))) letf))
 	 (setq letf (list 'let* (reverse letf) nil))
 	 (setf (caddr letf) (nth 3 exp))
 	 letf))))
@@ -676,35 +753,6 @@
 
 
 
-(defmacro cond (&body body)
-  (let (ret l)
-    (dolist (f body)
-      (if (cdr f)
-	  (progn
-	    (let ((cl
-		   `(if ,(car f)
-			,(if (cddr f)
-			     `(progn ,@(cdr f))
-			     (if (cdr f)
-				 (cadr f)))
-			,nil)))
-	      (if ret
-		  (setf (car l) cl l (cdddr cl))
-		  (setf ret cl l (cdddr cl)))))
-	  (progn
-	    (let* ((s (gensym))
-		   (cl
-		    `(let ((,s ,(car f)))
-		       (if ,s
-			   ,s
-			   ,nil))))
-	      (if ret
-		  (setf (car l) cl l (cddr (cdaddr cl)))
-		  (setf ret cl l (cddr (cdaddr cl))))))))
-    ret))
-
-
-
 (defmacro case (keyf &rest clauses)
   (let ((keysym (gensym))
 	outsym)
@@ -835,12 +883,21 @@
 
 
 (defun every (pred &rest sequences)
-  (let ((n (apply 'min (mapcar #'length sequences))))
-    (dotimes (i n)
-      (let ((args (mapcar (lambda (s) (elt s i)) sequences)))
+  (let ((seqs (copy-seq sequences)))
+    (do ((i 0 (+ 1 i)))
+	(nil)
+      (let ((args (maplist (lambda (s) (if (consp (car s))
+					   (prog1
+					       (caar s)
+					     (rplaca s (cdar s)))
+					   (if (null (car s))
+					       (return-from every t)
+					       (if (>= i (length (car s)))
+						   (return-from every t)
+						   (elt (car s) i)))))
+			   seqs)))
 	(if (not (apply pred args))
-	    (return-from every nil)))))
-  t)
+	    (return-from every nil))))))
 
 
 (defun some (pred &rest sequences)
@@ -857,23 +914,24 @@
 
 (defun member (obj list &key key test test-not)
   (let ((tst (or test
-	      (if test-not (complement test-not))
-	       #'eql)))
+		 (if test-not (complement test-not))
+		 #'eql)))
     (unless key
       (setq key #'identity))
-    (dotimes (i (length list))
-      (if (funcall tst obj (funcall key (car list)))
-	  (return-from member list))
-      (setq list (cdr list)))))
+    (do ((l list (cdr l)))
+	((not l) nil)
+      (if (funcall tst obj (funcall key (car l)))
+	  (return-from member l)))))
 
 
 (defun member-if (pred list &key key)
   (unless key
     (setq key #'identity))
-  (dotimes (i (length list))
-    (if (funcall pred (funcall key (car list)))
-	(return-from member-if list))
-    (setq list (cdr list))))
+  (do ((l list (cdr l)))
+      ((not l) nil)
+    (if (funcall pred (funcall key (car l)))
+	(return-from member-if l)))
+  nil)
 
 
 (defun member-if-not (pred list &key key)
@@ -888,16 +946,33 @@
 
 
 (defun find-if (pred seq &key from-end (start 0) end key)
-  (unless end
-    (setq end (length seq)))
   (unless key
     (setq key #'identity))
-  (dotimes (i (- end start))
-    (let ((j (if from-end
-		 (- end i 1)
-		 (+ start i))))
-      (if (funcall pred (funcall key (elt seq j)))
-	  (return-from find-if (elt seq j))))))
+  (if (consp seq)
+      (progn
+	(when from-end
+	  (setq seq (reverse seq))
+	  (let (tmp)
+	    (setq tmp start)
+	    (setq start (- (length seq) end))
+	    (setq end (- (length seq) tmp))))
+	(setq seq (nthcdr start seq))
+	(do ((cons seq (cdr cons))
+	     (i start (+ i 1)))
+	    ((or (not cons) (and end (>= i end))) nil)
+	  (if (funcall pred (funcall key (car cons)))
+	      (return-from find-if (car cons)))))
+      (if (not seq)
+	  nil
+	  (progn
+	    (unless end
+	      (setq end (length seq)))
+	    (dotimes (i (- end start))
+	      (let ((j (if from-end
+			   (- end i 1)
+			   (+ start i))))
+		(if (funcall pred (funcall key (elt seq j)))
+		    (return-from find-if (elt seq j)))))))))
 
 
 (defun find-if-not (pred seq &key from-end (start 0) end key)
@@ -1812,7 +1887,7 @@
 	(:import-from (setq imports (append imports (list (list (find-package (cadr opt)) (mapcar #'string (cddr opt)))))))
 	(:export (setq exports (append exports (mapcar #'string (cdr opt)))))
 	(:intern (setq intern (append intern (mapcar #'string (cdr opt)))))))
-    `(progn
+    `(eval-when (:compile-toplevel :load-toplevel :execute)
        (unless (find-package ,name)
 	 (make-package ,name))
        (rename-package ,name ,name ',nicks)
@@ -2017,8 +2092,9 @@
   (let (out)
     (do nil
 	(nil)
-      (if (member nil lists)
-	  (return nil))
+      (dolist (l lists)
+	(if (not l)
+	    (return-from maplist (reverse out))))
       (setq out (cons (apply fun lists) out))
       (setq lists (mapcar #'cdr lists)))
     (reverse out)))
@@ -2108,16 +2184,15 @@
 
 
 
-(deftype logical-pathname ()
-  'pathname)
-
-(defun logical-pathname (fn)
-  (pathname fn))
-
-
-(defun translate-logical-pathname (pn &key)
-  (pathname pn))
-
+(defun pathname-directory (fn &key case)
+  (let ((dir (cl-user:al-pathname-directory fn)))
+    (if dir
+	(let ((spl (split-pathname dir)))
+	  (if (string= (car spl) "/")
+	      (progn
+		(setf (car spl) :absolute)
+		spl)
+	      (cons :relative spl))))))
 
 (defun pathname-host (fn &key case))
 
@@ -2126,8 +2201,6 @@
 (defun pathname-version (fn &key case))
 
 
-(defun namestring (fn) (pathname-name fn))
-
 (defun file-namestring (fn) (pathname-name fn))
 
 (defun directory-namestring (fn))
@@ -2140,14 +2213,16 @@
 (defun merge-pathnames (pn &optional defpn defv)
   (let* ((p1 (pathname pn))
 	 (p2 (pathname defpn))
-	 (dir (or (pathname-directory p1) (pathname-directory p2)))
+	 (dir (or (cl-user:al-pathname-directory p1)
+		  (cl-user:al-pathname-directory p2)))
 	 (dirtype (if (and
 		       (> (length dir) 0) (char= (elt dir 0) #\/))
 		      :absolute
 		      :relative))
-	 (name (or (pathname-name p1) (pathname-name p2))))
+	 (name (or (pathname-name p1) (pathname-name p2)))
+	 (type (or (pathname-type p1) (pathname-type p2))))
     (make-pathname :directory (if dir (list dirtype dir))
-		   :name name)))
+		   :name name :type type)))
 
 
 
@@ -2159,6 +2234,137 @@
   (make-pathname :directory (cl-user:al-getenv "HOME")))
 
 
+(defun split-pathname (path)
+  (unless path
+    (return-from split-pathname nil))
+  (let ((pn (namestring (pathname path)))
+	out
+	start)
+    (dotimes (i (length pn))
+      (if (and (= i 0)
+	       (char= (elt pn 0) #\/))
+	  (setq out (cons (copy-seq "/") out))
+	  (if (not start)
+	      (when (not (char= (elt pn i) #\/))
+		(setq start i)
+		(if (= i (- (length pn) 1))
+		    (setq out (cons (subseq pn start) out))))
+	      (if (char= (elt pn i) #\/)
+		  (progn
+		    (setq out (cons (subseq pn start i) out))
+		    (setq start nil))
+		  (when (= i (- (length pn) 1))
+		    (setq out (cons (subseq pn start) out)))))))
+    (reverse out)))
+
+
+(defun pathname-match-p (path wild)
+  (let ((pdir (split-pathname (cl-user:al-pathname-directory path)))
+	(wdir (split-pathname (cl-user:al-pathname-directory wild))))
+    (while (and pdir wdir)
+      (when (string= (car wdir) "**")
+	(while (/= (length wdir) (length pdir))
+	  (setq pdir (cdr pdir)))
+	(setq wdir (cdr wdir))
+	(setq pdir (cdr pdir)))
+      (if (not (string= (car wdir) (car pdir)))
+	  (return-from pathname-match-p nil))
+      (setq pdir (cdr pdir))
+      (setq wdir (cdr wdir)))
+    (and
+     (not wdir)
+     (not pdir)
+     (or (string= (pathname-name path) (pathname-name wild))
+	 (string= (pathname-name wild) "*"))
+     (or (string= (pathname-type path) (pathname-type wild))
+	 (string= (pathname-name wild) "*")
+	 (string= (pathname-type wild) "*")))))
+
+
+
+(defun ensure-directory (path)
+  (let ((path (namestring path)))
+    (if (char= (elt path (1- (length path))) #\/)
+	path
+	(concatenate 'string path "/"))))
+
+
+(defun cl-user::al-list-directory-with-full-path (dir)
+  (mapcar (lambda (f) (merge-pathnames (ensure-directory dir)
+				       f))
+	  (cl-user:al-list-directory dir)))
+(export 'cl-user::al-list-directory-with-full-path 'cl-user)
+
+
+(defun exclude-directory (path)
+  "Returns the pathname without the directory, that is name+type (or just the name in Posix parlance)."
+  (let ((name (pathname-name path))
+	(type (pathname-type path)))
+    (if type
+	(concatenate 'string name "." type)
+	(or name (make-string 0)))))
+
+
+(defun cl-user::al-list-subdirs (dir)
+  "List the subdirectories of DIR with full path.  Excludes '.' and '..'."
+  (let (out
+	(dir (namestring dir)))
+    (dolist (f (cl-user:al-list-directory-with-full-path dir))
+      (if (and (string/= (exclude-directory f) ".")
+	       (string/= (exclude-directory f) "..")
+	       (cl-user:al-directoryp f))
+	  (setq out (cons (ensure-directory f) out))))
+    out))
+(export 'cl-user::al-list-subdirs 'cl-user)
+
+
+(defun directory (fn)
+  (let ((name (pathname-name fn))
+	(type (pathname-type fn))
+	(dir (cl-user:al-pathname-directory fn))
+	(full (namestring fn))
+	(spl (split-pathname fn)))
+    (cond
+      ((and (string= name "*")
+	    (or (not type)
+		(string= type "*")))
+       (cl-user:al-list-directory-with-full-path (or dir ".")))
+      ((string= name "*")
+       (remove-if (lambda (f) (string/= (or (pathname-type f) "") type))
+		  (cl-user:al-list-directory-with-full-path (or dir "."))))
+      ((string= (car (last spl)) "*")
+       (mapcar 'pathname
+	       (cl-user:al-list-subdirs (subseq full 0 (- (length full) 2)))))
+      (t
+       (if (probe-file fn)
+	   (list (pathname fn)))))))
+
+
+(defun translate-pathname (src from-wild dest &key)
+  (let ((srcdir (pathname-directory src))
+	(destdir (pathname-directory dest))
+	(destname (pathname-name dest))
+	(desttype (pathname-type dest))
+	outdir prevcons)
+    (if destdir
+	(do ((d (cdr destdir) (cdr d)))
+	    ((not d))
+	  (when (or (string= (car d) "*")
+		    (string= (car d) "**"))
+	    (if prevcons
+		(setf (cdr prevcons) (cdr srcdir))
+		(setq destdir srcdir))
+	    (setf (cdr (last srcdir)) (cdr d)))
+	  (setq prevcons d))
+	(setq destdir srcdir))
+    (make-pathname :directory destdir
+		   :name (if (or (not destname) (string= destname "*"))
+			     (pathname-name src)
+			     destname)
+		   :type (if (or (not desttype) (string= desttype "*"))
+			     (pathname-type src)
+			     desttype))))
+
 
 
 (defmacro with-open-file ((str fn &rest opts) &body forms)
@@ -2379,84 +2585,96 @@
 
 
 (defun loop-parse-iteration (forms)
-  (let (iter)
-    (setq iter `(,(cadr forms)))
+  (let (var iter)
+    (setq var (cadr forms))
     (setq forms (cddr forms))
+    (if (member (car forms) '(fixnum float t nil))
+	(setq forms (cdr forms))
+	(if (and (symbolp (car forms))
+		 (string= (car forms) "OF-TYPE"))
+	    (setq forms (cddr forms))))
     (do ((f forms))
 	((not f) (setq forms f))
       (if (atom (car f))
 	  (let ((sym (string (car f))))
 	    (cond
 	      ((string= sym "FROM")
-	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
-	       (setf (elt iter 3) (cadr f))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
+	       (setf (elt iter 2) (cadr f))
+	       (setf (elt iter 4) '>)
 	       (setq f (cddr f)))
 	      ((string= sym "DOWNFROM")
-	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
-	       (setf (elt iter 3) (cadr f))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
+	       (setf (elt iter 2) (cadr f))
+	       (setf (elt iter 4) '<)
 	       (setf (elt iter 5) -1)
 	       (setq f (cddr f)))
 	      ((string= sym "UPFROM")
-	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
-	       (setf (elt iter 3) (cadr f))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
+	       (setf (elt iter 2) (cadr f))
+	       (setf (elt iter 4) '>)
 	       (setf (elt iter 5) 1)
 	       (setq f (cddr f)))
 	      ((string= sym "TO")
-	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
-	       (setf (elt iter 4) (cadr f))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
+	       (setf (elt iter 3) (cadr f))
 	       (setq f (cddr f)))
 	      ((string= sym "DOWNTO")
-	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
-	       (setf (elt iter 4) (cadr f))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
+	       (setf (elt iter 3) (cadr f))
+	       (setf (elt iter 4) '<)
 	       (setf (elt iter 5) -1)
 	       (setq f (cddr f)))
 	      ((string= sym "UPTO")
-               (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
-	       (setf (elt iter 4) (cadr f))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
+	       (setf (elt iter 3) (cadr f))
+	       (setf (elt iter 4) '>)
 	       (setf (elt iter 5) 1)
 	       (setq f (cddr f)))
 	      ((string= sym "BELOW")
-	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
-	       (setf (elt iter 4) (1- (cadr f)))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
+	       (setf (elt iter 3) (cadr f))
+	       (setf (elt iter 4) '>=)
 	       (setq f (cddr f)))
 	      ((string= sym "ABOVE")
-	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
-	       (setf (elt iter 4) (1+ (cadr f)))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
+	       (setf (elt iter 3) (cadr f))
+	       (setf (elt iter 4) '<=)
 	       (setq f (cddr f)))
 	      ((string= sym "BY")
-	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
+	       (unless iter
+		 (setq iter `(:from ,var 0 nil > ,1)))
 	       (cond
 		 ((eq (car iter) :from)
-		  (setf (elt iter 6) (cadr f)))
+		  (setf (elt iter 5) (* (elt iter 5) (cadr f))))
 		 ((member (car iter) '(:in :on))
 		  (setf (elt iter 4) (cadr f))))
 	       (setq f (cddr f)))
 	      ((string= sym "IN")
-	       (setq iter `(:in ,(car iter) nil ,(cadr f) ,#'cdr))
+	       (setq iter `(:in ,var nil ,(cadr f) ,#'cdr))
 	       (setq f (cddr f)))
 	      ((string= sym "ON")
-	       (setq iter `(:on ,(car iter) nil ,(cadr f) ,#'cdr))
+	       (setq iter `(:on ,var nil ,(cadr f) ,#'cdr))
 	       (setq f (cddr f)))
 	      ((string= sym "=")
-	       (setq iter `(:eq ,(car iter) ,(cadr f) ,nil))
+	       (setq iter `(:eq ,var ,(cadr f) ,nil))
 	       (setq f (cddr f)))
 	      ((string= sym "THEN")
 	       (setf (elt iter 3) (cadr f))
 	       (setq f (cddr f)))
 	      ((string= sym "ACROSS")
-	       (setq iter `(:across ,(car iter) nil nil ,(cadr f)))
+	       (setq iter `(:across ,var nil nil ,(cadr f)))
 	       (setq f (cddr f)))
 	      ((string= sym "BEING")
-	       (setq iter `(:hashtable ,(car iter) nil nil ,nil))
+	       (setq iter `(:hashtable ,var nil nil ,nil))
 	       (setq f (cdr f))
 	       (if (or (string= (string (car f)) "EACH")
 		       (string= (string (car f)) "THE"))
@@ -2485,7 +2703,7 @@
 	  (progn
 	    (setq forms f)
 	    (return nil))))
-    (if (member (car iter) '(:from :in :on :across))
+    (if (member (car iter) '(:in :on :across))
 	(setf (elt iter 2) (gensym)))
     (if (eq (car iter) :across)
 	(setf (elt iter 3) (gensym)))
@@ -2756,22 +2974,20 @@
 				  (cl-user:al-loopy-setq ,(elt i 2) (funcall ,(elt i 4) ,(elt i 2)))
 				  . ,iteration-setup)))
 	(when (eq (car i) :from)
-	  (setq vrs (cons (elt i 1) vrs))
-	  (setq initial-setup `(,@(if (elt i 4)
-				      (list `(when ,(if (= (elt i 5) 1)
-							`(> ,(elt i 1) ,(elt i 4))
-							`(< ,(elt i 1) ,(elt i 4)))
-					       (setq ,skip-body-sym t)
-					       (go ,prologue-tag))))
-				  (setq ,(elt i 1) ,(elt i 3))
-				  . ,initial-setup))
-	  (setq iteration-setup `(,@(if (elt i 4)
-					(list `(if ,(if (= (elt i 5) 1)
-							`(> ,(elt i 1) ,(elt i 4))
-							`(< ,(elt i 1) ,(elt i 4)))
-						   (return-from ,inner-block-name nil))))
-				    (setq ,(elt i 1) (+ ,(elt i 1) ,(* (elt i 5) (elt i 6))))
-				    . ,iteration-setup)))
+	  (let ((stopvar (gensym))
+		(stepvar (gensym)))
+	    (setq vrs (list* (elt i 1) (list stopvar (elt i 3)) (list stepvar (elt i 5)) vrs))
+	    (setq initial-setup `(,@(if (elt i 3)
+					(list `(when (,(elt i 4) ,(elt i 1) ,stopvar)
+						 (setq ,skip-body-sym t)
+						 (go ,prologue-tag))))
+				    (setq ,(elt i 1) ,(elt i 2))
+				    . ,initial-setup))
+	    (setq iteration-setup `(,@(if (elt i 3)
+					  (list `(if (,(elt i 4) ,(elt i 1) ,stopvar)
+						     (return-from ,inner-block-name nil))))
+				      (setq ,(elt i 1) (+ ,(elt i 1) ,stepvar))
+				      . ,iteration-setup))))
 	(when (eq (car i) :eq)
 	  (setq vrs (cons (elt i 1) vrs))
 	  (setq initial-setup `((cl-user:al-loopy-setq ,(elt i 1) ,(elt i 2))
@@ -2920,13 +3136,122 @@
 
 
 
+(defparameter *readtable* nil)
+
+
+
 (defmacro with-compilation-unit (opts &rest forms)
   `(progn
      ,@forms))
 
 
 
-(defparameter *readtable* nil)
+(defparameter *compile-file-truename* nil)
+
+(defparameter *compile-file-pathname* nil)
+
+(defparameter *compile-print* nil)
+
+(defparameter *compile-verbose* nil)
+
+
+(defun compile-file-pathname (infile &key output-file &allow-other-keys)
+  (let ((dir (cl-user:al-pathname-directory infile)))
+    (make-pathname :directory (and dir (if (char= (elt dir 0) #\/) dir (list :relative dir)))
+		   :name (pathname-name infile)
+		   :type "alc")))
+
+
+(defun write-preserving-gensyms (obj str gensyms)
+  (typecase obj
+    (cons
+     (write-string "(" str)
+     (let ((firstobjp t))
+       (do ((c obj (cdr c)))
+	   (nil)
+	 (unless firstobjp
+	   (write-string " " str))
+	 (setq firstobjp nil)
+	 (setq gensyms (write-preserving-gensyms (car c) str gensyms))
+	 (if (not (cdr c))
+	     (return nil))
+	 (when (atom (cdr c))
+	   (write-string " . " str)
+	   (setq gensyms (write-preserving-gensyms (cdr c) str gensyms))
+	   (return nil)))
+       (write-string ")" str)))
+    (symbol
+     (let ((cl-user:*al-print-always-two-colons* t))
+       (if (symbol-package obj)
+	   (write obj :stream str)
+	   (let ((ind (position obj gensyms :test 'eq)))
+	     (if ind
+		 (format str "#~s#" ind)
+		 (progn
+		   (setq gensyms (nconc gensyms (list obj)))
+		   (format str "#~s=~s" (1- (length gensyms)) obj)))))))
+    (cl-user:al-backquote
+     (write-string "`" str)
+     (setq gensyms (write-preserving-gensyms (cl-user:al-next obj) str gensyms)))
+    (cl-user:al-comma
+     (write-string "," str)
+     (setq gensyms (write-preserving-gensyms (cl-user:al-next obj) str gensyms)))
+    (cl-user:al-at
+     (write-string "@" str)
+     (setq gensyms (write-preserving-gensyms (cl-user:al-next obj) str gensyms)))
+    (cl-user:al-dot
+     (write-string "." str)
+     (setq gensyms (write-preserving-gensyms (cl-user:al-next obj) str gensyms)))
+    (otherwise (write obj :stream str)))
+  gensyms)
+
+
+(defun parse-toplevel-form-at-compile-time (form)
+  (if (consp form)
+      (cond
+	((eq (car form) 'progn)
+	 (dolist (f (cdr form))
+	   (parse-toplevel-form-at-compile-time f)))
+	((member (car form) '(in-package defmacro) :test #'eq)
+	 (eval form))
+	((and
+	  (eq (car form) 'eval-when)
+	  (member :compile-toplevel (cadr form) :test #'eq))
+	 (eval `(progn
+		  ,@(cddr form)))))))
+
+
+(defun compile-file (infile &key (output-file infile) (verbose *compile-verbose*) (print *compile-print*) external-format)
+  (let* ((*compile-file-truename* infile)
+	 (*compile-file-pathname* infile)
+	 (outname (compile-file-pathname output-file))
+	 (old-readtable *readtable*)
+	 (old-package *package*)
+	 (eofsym (gensym)))
+    (unwind-protect
+	 (progn
+	   (if verbose
+	       (format t ";;; Compiling file ~a...~%" infile))
+	   (with-open-file (instr infile :direction :input)
+	     (with-open-file (outstr outname :direction :output :if-exists :overwrite)
+	       (do nil
+		   (nil)
+		 (let ((obj (read instr nil eofsym)))
+		   (if (eq obj eofsym)
+		       (return-from compile-file (values outname nil nil)))
+		   (when print
+		     (format t "Compiling ")
+		     (write obj)
+		     (terpri))
+		   (setq obj (cl-user:al-compile-form obj))
+		   (parse-toplevel-form-at-compile-time obj)
+		   (if print
+		       (terpri))
+		   (write-preserving-gensyms obj outstr nil)
+		   (terpri outstr)
+		   (terpri outstr))))))
+      (setq *readtable* old-readtable)
+      (setq *package* old-package))))
 
 
 
@@ -3045,6 +3370,10 @@
 
 
 
+(defun muffle-warning (&optional cond))
+
+
+
 (defgeneric documentation (x doc-type))
 
 (defmethod documentation ((x t) (doc-type t)))
@@ -3103,18 +3432,25 @@
 	  stringp simple-string-p bit-vector-p simple-bit-vector-p hash-table-p
 	  pathnamep streamp realp numberp check-type assert macroexpand equal
 	  fdefinition complement mapc mapcan maplist mapl mapcon map-into reduce
-	  merge logical-pathname translate-logical-pathname pathname-host
-	  pathname-device pathname-version namestring file-namestring
-	  directory-namestring host-namestring enough-namestring merge-pathnames
-	  file-author file-write-date user-homedir-pathname with-open-file
-	  terpri write-line write-sequence prin1 princ print write-to-string
-	  prin1-to-string princ-to-string force-output with-input-from-string
-	  with-output-to-string pprint do-all-symbols find-all-symbols
-	  with-slots with-accessors loop format encode-universal-time
-	  with-compilation-unit *readtable* with-standard-io-syntax handler-case
-	  restart-case with-simple-restart find-restart cerror break
-	  ignore-errors abort continue documentation))
+	  merge pathname-directory pathname-host pathname-device
+	  pathname-version file-namestring directory-namestring host-namestring
+	  enough-namestring merge-pathnames file-author file-write-date
+	  user-homedir-pathname pathname-match-p directory translate-pathname
+	  with-open-file terpri write-line write-sequence prin1 princ print
+	  write-to-string prin1-to-string princ-to-string force-output
+	  with-input-from-string with-output-to-string pprint do-all-symbols
+	  find-all-symbols with-slots with-accessors loop format
+	  encode-universal-time *readtable* with-compilation-unit
+	  *compile-file-truename* *compile-file-pathname* *compile-print*
+	  *compile-verbose* compile-file-pathname compile-file
+	  with-standard-io-syntax handler-case restart-case with-simple-restart
+	  find-restart cerror break ignore-errors abort continue muffle-warning
+	  documentation))
 
 
 
 (in-package cl-user)
+
+
+
+(setq cl-user:*al-compile-when-defining* nil)
